@@ -61,13 +61,15 @@ description. Methods are listed in alphabetical order.
 #
 package XAO::DO::FS::Hash;
 use strict;
+use utf8;
 use XAO::Utils;
 use XAO::Objects;
+use Encode;
 
 use base XAO::Objects->load(objname => 'FS::Glue');
 
 use vars qw($VERSION);
-$VERSION=(0+sprintf('%u.%03u',(q$Id: Hash.pm,v 2.1 2005/01/14 00:23:54 am Exp $ =~ /\s(\d+)\.(\d+)\s/))) || die "Bad VERSION";
+$VERSION=(0+sprintf('%u.%03u',(q$Id: Hash.pm,v 2.10 2008/09/14 07:07:12 am Exp $ =~ /\s(\d+)\.(\d+)\s/))) || die "Bad VERSION";
 
 ###############################################################################
 
@@ -91,6 +93,26 @@ Only makes sense for 'list' type - sets the name of class for the
 objects that would be contained in the list. That class have to exist
 and be available for object loader at the time of the call.
 
+=item charset
+
+For fields of 'text' type this can be one of 'binary', 'latin1', or
+'utf8'. For compatibility with older version that did not have a
+dedicated 'blob' data type the default charset is 'binary'.
+
+Fields of 'latin1' or 'utf8' charset have no guaranteed behaviour on
+characters not defined for these charsets. Such characters can be
+skipped or converted to '?' for instance.
+
+'Binary' charset fields will accept any data just like 'blob' data type,
+but the sorting order and word matching in search may differ from older
+versions. As another side effect, searches on 'binary' charset fields
+are case sensitive.
+
+B<NOTE:> The 'utf8' charset only supports the Basic Multilingual Plane
+characters (0x0000-0xffff). These characters in UTF-8 encoding are
+guaranteed to take up from 1 to 3 bytes. Supplemental Unicode planes are
+not supported.
+
 =item connector
 
 Optional name of the key that would refer objects in the list to the
@@ -115,13 +137,12 @@ going to use this field in searches a lot. This works especially good on
 `real' and `integer' fields when you later search on them using ranges
 or equivalence.
 
-For text searches it is usually better to make a field of type `words'
-then making an index on it. Indexing text fields works great mostly on
-'eq' and 'ne' search operators, not 'ws' or 'wq'.
+Note, that indexing text fields works mostly on 'eq', 'ne', and 'sw'
+search operators, not 'cs', 'ws', or 'wq'.
 
-Regardless of your use of indexes searches are guaranteed to produc
-equal results. Indexing can only improve performance or decrease it when
-used incorrectly.
+Regardless of your use of indexes searches are guaranteed to produce
+equal results. Indexing can only improve performance (or decrease it when
+used incorrectly).
 
 =item key
 
@@ -132,6 +153,11 @@ When possible it is recommended to name the key as the class name or
 some derivative from it with '_id' suffix. If you add Orders list
 placeholder it is recommended to call the key 'order_id'.
 
+=item key_charset
+
+Charset to be used for the list key. Default is 'binary' which should be
+sufficient for the majority of applications.
+
 =item key_format
 
 Only for 'list' placeholders. Specifies format of the auto-generated
@@ -149,7 +175,6 @@ function. It never evaluates to 'false' in perl sense, always starts
 from a letter and has some other constraints - see L<XAO::Utils> for
 details.
 
-
 =item <$AUTOINC$>
 
 This is an auto-incrementing unsigned at least 32-bit integer value. It
@@ -189,9 +214,13 @@ name length in Hashes for consistency reasons).
 
 =item maxlength
 
-Maximum length for 'text' type in characters. If later on you will try
-to store longer text string into that field an error will be
-thrown. Default is 100.
+Maximum length for 'text' type in bytes (so for 'utf8' charset you may
+store less characters than the maxlength depending on the data). If
+later on you will try to store longer text string into that field an
+error will be thrown.
+
+Default is 100, but using default length is deprecated -- it may become
+illegal in future versions.
 
 =item maxvalue
 
@@ -210,6 +239,13 @@ inclusive. Default is -2147483648 for integer property. If you use
 positive value or zero for integer property you're converting that
 property to unsigned integer value effectively.
 
+=item scale
+
+Only makes sense for 'real' type. When set the field is stored as a
+fixed precision decimal with guaranteed math properties (as opposed to
+plain 'real' values where there may not be a way to store exactly 3.0
+for example).
+
 =item name
 
 Placeholder name -- this is the name you would then use in put() and
@@ -228,7 +264,7 @@ be created from class name (something like 'osCustomer_Address' for
 
 =item type
 
-Placeholder type, one of 'text', 'words', 'integer', 'unsigned', 'real'
+Placeholder type, one of 'text', 'blob', 'integer', 'unsigned', 'real'
 or 'list'.
 
 =item unique
@@ -267,8 +303,8 @@ sub add_placeholder ($%) {
 	my $self=shift;
 	my $args=get_args(\@_);
 
-	my $name=$args->{name} || throw $self 'add_placeholder - no name given';
-	my $type=$args->{type} || throw $self 'add_placeholder - no type given';
+	my $name=$args->{'name'} || throw $self 'add_placeholder - no name given';
+	my $type=$args->{'type'} || throw $self 'add_placeholder - no type given';
 
 	$self->check_name($name) ||
 	    throw $self "add_placeholder - bad name ($name)";
@@ -326,20 +362,47 @@ sub build_structure ($%) {
     my $self=shift;
     my $args=get_args(\@_);
 
+    my %changed_charset;
+    my %changed_scale;
+
+    $self->_consistency_checked ||
+        throw $self "- can't build_structure without 'check_consistency' on loading";
+
     foreach my $name (keys %$args) {
         my %ph;
         @ph{keys %{$args->{$name}}}=values %{$args->{$name}};
-        $ph{name}=$name;
+        $ph{'name'}=$name;
 
         my $desc=$self->describe($name);
         if($desc) {
             while(my ($n,$v)=each %ph) {
                 next if $n eq 'name';
                 next if $n eq 'structure';
+                next if !defined $v;
+
                 my $dbv=$desc->{$n};
-                (!defined($v) && !defined($dbv)) || $dbv eq $v ||
-                    throw $self "build_structure - structure mismatch, " .
-                                "property=$name, ($n,$v) <> ($n,$dbv)";
+
+                if($n eq 'charset') {
+                    $v eq 'binary' || Encode::resolve_alias($v) ||
+                        throw $self "- new charset '$v' is not supported for field '$name'";
+
+                    if($v ne ($dbv || '')) {
+                        $changed_charset{$name}=$v;
+                    }
+                }
+                elsif($n eq 'scale') {
+                    $dbv||=0;
+
+                    if($v<$dbv) {
+                        throw $self "- new scale '$v' is lower than existing '$dbv' for '$name' (only upwards scale changes are automatic)";
+                    }
+                    elsif($v>$dbv) {
+                        $changed_scale{$name}=$v;
+                    }
+                }
+                elsif(!defined $dbv || (defined $dbv && $dbv ne $v)) {
+                    throw $self "- structure mismatch, property=$name, ($n,$v) <> ($n,".(defined $dbv ? $dbv : '<undef>').")";
+                }
             }
         }
         else {
@@ -348,12 +411,38 @@ sub build_structure ($%) {
 
         # Building recursive structures
         #
-        if($ph{type} eq 'list' && $ph{structure}) {
-            my $ro=XAO::Objects->new(objname => $ph{class},
+        if($ph{'type'} eq 'list' && $ph{'structure'}) {
+            my $ro=XAO::Objects->new(objname => $ph{'class'},
                                      glue => $self->_glue);
-            $ro->build_structure($ph{structure});
+            $ro->build_structure($ph{'structure'});
         }
     }
+
+    if(keys %changed_charset) {
+        my $debug_status=XAO::Utils::get_debug();
+        XAO::Utils::set_debug(1);
+
+        dprint "Some text fields in ".$$self->{'class'}." have changed charset values.";
+        dprint "An automatic conversion will be attempted in 10 seconds. Interrupt to abort.";
+        sleep 10;
+        dprint "Converting...";
+        $self->_charset_change(\%changed_charset);
+
+        XAO::Utils::set_debug($debug_status);
+    }
+
+    if(keys %changed_scale) {
+        my $debug_status=XAO::Utils::get_debug();
+        XAO::Utils::set_debug(1);
+
+        dprint "Some 'real' fields in ".$$self->{'class'}." have changed scale values.";
+        dprint "An automatic conversion will be attempted in 10 seconds. Interrupt to abort.";
+        sleep 10;
+        dprint "Converting...";
+        $self->_scale_change(\%changed_scale);
+
+        XAO::Utils::set_debug($debug_status);
+    }
 }
 
 ###############################################################################
@@ -370,7 +459,7 @@ Will return undef on detached objects.
 
 sub collection_key ($) {
     my $self=shift;
-    return $$self->{unique_id};
+    return $$self->{'unique_id'};
 }
 
 ###############################################################################
@@ -408,7 +497,7 @@ sub container_object ($) {
 
     return undef if $self->objname eq 'FS::Global';
 
-    my $list_base_name=$$self->{list_base_name} ||
+    my $list_base_name=$$self->{'list_base_name'} ||
         $self->throw("container_object - the object was not retrieved from the database");
 
     ##
@@ -450,7 +539,7 @@ sub defined ($$) {
 
     return 1 if $name eq 'unique_id';
 
-    my $f=$self->_class_description->{fields};
+    my $f=$self->_class_description->{'fields'};
 
     exists $f->{$name} ||
         throw $self "defined - unknown property ($name)";
@@ -482,7 +571,7 @@ sub delete ($$) {
     my $field=$self->_field_description($name);
     $field || $self->throw("delete($name) - unknown field name");
 
-    my $type=$field->{type};
+    my $type=$field->{'type'};
     if($type eq 'list') {
         $self->get($name)->destroy();
     }
@@ -514,8 +603,8 @@ non-existing fields.
 Example:
 
  my $description=$customer->describe('first_name');
- print "Type: $description->{type}\n";
- print "Maximum length: $description->{maxlength}\n";
+ print "Type: $description->{'type'}\n";
+ print "Maximum length: $description->{'maxlength'}\n";
 
 =cut
 
@@ -627,7 +716,7 @@ sub drop_placeholder ($$) {
     my $field=$self->_field_description($name) ||
         $self->throw("drop_placeholder - placeholder does not exist ($name)");
 
-	if($field->{type} eq 'list') {
+	if($field->{'type'} eq 'list') {
 	    $self->_drop_list_placeholder($name);
 	} else {
 	    $self->_drop_data_placeholder($name);
@@ -653,7 +742,7 @@ sub exists ($$) {
     my $self=shift;
     my $name=shift;
     return 1 if $name eq 'unique_id';
-    my $f=$self->_class_description->{fields};
+    my $f=$self->_class_description->{'fields'};
     exists $f->{$name};
 }
 
@@ -726,12 +815,12 @@ sub get ($$) {
         my $field=$self->_field_description($name) ||
                     $self->throw("get - unknown property ($name)");
 
-        my $type=$field->{type};
-        if($$self->{detached}) {
+        my $type=$field->{'type'};
+        if($$self->{'detached'}) {
             if($type eq 'list') {
                 $self->throw("get - retrieval of lists on detached objects not implemented yet");
             } else {
-                my $value=$$self->{data}->{$name};
+                my $value=$$self->{'data'}->{$name};
                 $value=$self->_field_default($name,$field) unless defined($value);
                 return $value;
             }
@@ -742,15 +831,27 @@ sub get ($$) {
                             objname => 'FS::List',
                             glue => $self->_glue,
                             uri => $self->uri($name),
-                            class_name => $field->{class},
-                            base_name => $$self->{objname},
-                            base_id => $$self->{unique_id},
+                            class_name => $field->{'class'},
+                            base_name => $$self->{'objname'},
+                            base_id => $$self->{'unique_id'},
                             key_value => $name);
             }
             else {
                 my $value=$self->_retrieve_data_fields($name);
+
                 defined $value ||
                     throw $self "get('$name') - db query returned undef";
+
+                if($type eq 'text') {
+                    my $charset=$field->{'charset'};
+                    if($charset ne 'binary' && !Encode::is_utf8($value)) {
+                        $value=Encode::decode($charset,$value,Encode::FB_DEFAULT);
+                    }
+                }
+                elsif($type eq 'real' && !(0+$value)) {
+                    $value=0;
+                }
+
                 return $value;
             }
         }
@@ -761,20 +862,20 @@ sub get ($$) {
     #
     else {
 
-        my $fields=$self->_class_description->{fields} ||
+        my $fields=$self->_class_description->{'fields'} ||
             $self->throw("get - internal data problem");
 
         my @datanames=map {
             my $f=$fields->{$_} || $self->throw("get - unknown property ($_)");
-            $f->{type} eq 'list' ? () : ($_);
+            $f->{'type'} eq 'list' ? () : ($_);
         } @_;
 
-        if($$self->{detached}) {
+        if($$self->{'detached'}) {
             scalar(@datanames) == scalar(@_) ||
                 $self->throw("get - retrieval of lists on detached objects not implemented yet");
 
             return map {
-                my $value=$$self->{data}->{$_};
+                my $value=$$self->{'data'}->{$_};
                 if(!defined($value)) {
                     $value=$self->_field_default($_,$fields->{$_});
                 }
@@ -783,25 +884,39 @@ sub get ($$) {
         }
         else {
             my %datahash;
+
             if(@datanames) {
                 @datahash{@datanames}=$self->_retrieve_data_fields(@datanames);
             }
 
             return map {
-                if($fields->{$_}->{type} eq 'list') {
+                my $type=$fields->{$_}->{'type'};
+
+                if($type eq 'list') {
                     XAO::Objects->new(objname => 'FS::List',
                                       glue => $self->_glue,
                                       uri => $self->uri($_),
-                                      class_name => $fields->{$_}->{class},
-                                      base_name => $$self->{objname},
-                                      base_id => $$self->{unique_id},
+                                      class_name => $fields->{$_}->{'class'},
+                                      base_name => $$self->{'objname'},
+                                      base_id => $$self->{'unique_id'},
                                       key_value => $_);
                 }
                 else {
                     my $value=$datahash{$_};
-                    if(!defined($value)) {
-                        $value=$self->_field_default($_,$fields->{$_});
+
+                    defined $value ||
+                        throw $self "get('$_') - db query returned undef";
+
+                    if($type eq 'text') {
+                        my $charset=$fields->{$_}->{'charset'};
+                        if($charset ne 'binary' && !Encode::is_utf8($value)) {
+                            $value=Encode::decode($charset,$value,Encode::FB_DEFAULT);
+                        }
                     }
+                    elsif($type eq 'real' && !(0+$value)) {
+                        $value=0;
+                    }
+
                     $value;
                 }
             } @_;
@@ -831,7 +946,7 @@ false otherwise.
 
 sub is_attached ($) {
     my $self=shift;
-    ! $$self->{detached};
+    ! $$self->{'detached'};
 }
 
 ###############################################################################
@@ -853,10 +968,10 @@ Example:
 sub keys ($) {
     my $self=shift;
 
-    my $fields=$self->_class_description()->{fields};
+    my $fields=$self->_class_description()->{'fields'};
     my @list;
     foreach my $key (keys %{$fields}) {
-        next if $fields->{$key}->{type} eq 'connector';
+        next if $fields->{$key}->{'type'} eq 'connector';
         push(@list,$key);
     }
 
@@ -895,30 +1010,30 @@ sub new ($%) {
     my $self=$class->SUPER::new(@_);
     my $args=get_args(\@_);
 
-    $$self->{unique_id}=$args->{unique_id};
+    $$self->{'unique_id'}=$args->{'unique_id'};
 
-    if(! $$self->{unique_id}) {
-        $$self->{detached}=1;
+    if(! $$self->{'unique_id'}) {
+        $$self->{'detached'}=1;
     }
     else {
-        $$self->{key_name}=$args->{key_name};
-        defined($$self->{key_name}) || $self->throw("new - no 'key_name' passed");
-        $$self->{key_value}=$args->{key_value} ||
-                            $self->get($args->{key_name});
+        $$self->{'key_name'}=$args->{'key_name'};
+        defined($$self->{'key_name'}) || $self->throw("new - no 'key_name' passed");
+        $$self->{'key_value'}=$args->{'key_value'} ||
+                            $self->get($args->{'key_name'});
 
-        $$self->{list_base_name}=$args->{list_base_name};
-        $$self->{list_base_id}=$args->{list_base_id};
-        $$self->{list_key_value}=$args->{list_key_value};
+        $$self->{'list_base_name'}=$args->{'list_base_name'};
+        $$self->{'list_base_id'}=$args->{'list_base_id'};
+        $$self->{'list_key_value'}=$args->{'list_key_value'};
 
-        if(! defined($$self->{uri})) {
-            my $uri=$$self->{key_value};
+        if(! defined($$self->{'uri'})) {
+            my $uri=$$self->{'key_value'};
             my $p=$self;
             while(defined($p=$p->container_object)) {
                 my $ck=$p->container_key();
                 $ck='' unless defined($ck);
                 $uri=$ck . '/' . $uri;
             }
-            $$self->{uri}=$uri;
+            $$self->{'uri'}=$uri;
         }
     }
 
@@ -997,58 +1112,83 @@ sub put ($$$) {
     my $self=shift;
     my $data=get_args(\@_);
 
-    my $detached=$$self->{detached};
+    my $detached=$$self->{'detached'};
 
     my @data_keys=CORE::keys %$data;
     foreach my $name (@data_keys) {
         my $value=$data->{$name};
 
         my $field=$self->_field_description($name);
-        $field || $self->throw("put($name,...) - not defined field name");
+        $field || $self->throw("- {{INTERNAL:Not defined field name '$name'}}");
 
         if(!defined $value) {
             $data->{$name}=$value=$self->_field_default($name,$field);
         }
 
-        my $type=$field->{type};
+        my $type=$field->{'type'};
         if($type eq 'list') {
-            $self->throw("put - storing lists not implemented yet");
+            $self->throw("- {{INTERNAL:Storing lists not implemented yet}}");
         }
         elsif($type eq 'key') {
-            $self->throw("put - attempt to modify hash key");
+            $self->throw("- {{INTERNAL:Attempt to modify hash key}}");
         }
         elsif($name eq 'unique_id') {
-            $self->throw("put - attempt to modify unique_id");
+            $self->throw("- {{INTERNAL:Attempt to modify unique_id}}");
         }
-        elsif($type eq 'text' || $type eq 'words') {
-            length($value) <= $field->{maxlength} ||
-                $self->throw("put - value is longer then $field->{maxlength} for $name");
+        elsif($type eq 'blob') {
+            if(Encode::is_utf8($value)) {
+                $data->{$name}=$value=Encode::encode('utf8',$value);
+            }
+            length($value) <= $field->{'maxlength'} ||
+                $self->throw("- {{INPUT:Value is longer than $field->{'maxlength'} for '$name'}}");
+        }
+        elsif($type eq 'text') {
+            if(Encode::is_utf8($value)) {
+                length($value) <= $field->{'maxlength'} ||
+                    $self->throw("- {{INPUT:Value is longer than $field->{'maxlength'} for '$name'}}");
+
+                my $charset=$field->{'charset'};
+
+                if($value=~/[\x{10000}-\x{ffffffff}]/ && $charset eq 'utf8') {
+                    $self->throw("- {{INPUT:Unsupported supplemental unicode value for '$name'}}");
+                }
+
+                $data->{$name}=$value=Encode::encode($charset eq 'binary' ? 'utf8' : $charset,$value);
+            }
+            if($field->{'charset'} eq 'binary') {
+                length($value) <= $field->{'maxlength'} ||
+                    $self->throw("- {{INPUT:Value is longer than $field->{'maxlength'} for '$name'}}");
+            }
+            else {
+                # Checking characters length with binary source is
+                # too expensive, not doing it This should be a rare
+                # situation anyway, and MySQL will chop off the extra if
+                # needed.
+            }
         }
         elsif($type eq 'integer' || $type eq 'real') {
-            $value=$self->_field_default($name,$field) if $value eq '';
+            $data->{$name}=$value=$self->_field_default($name,$field) if $value eq '';
 
-            !defined($field->{minvalue}) || $value>=$field->{minvalue} ||
-                $self->throw("put - value ($value) is less then $field->{minvalue} for $name");
+            !defined($field->{'minvalue'}) || $value>=$field->{'minvalue'} ||
+                $self->throw("- {{INPUT:Value ($value) is less then $field->{'minvalue'} for $name}}");
 
-            !defined($field->{maxvalue}) || $value<=$field->{maxvalue} ||
-                $self->throw("put - value ($value) is bigger then $field->{maxvalue} for $name");
+            !defined($field->{'maxvalue'}) || $value<=$field->{'maxvalue'} ||
+                $self->throw("- {{INPUT:Value ($value) is bigger then $field->{'maxvalue'} for $name}}");
         }
         else {
-            throw $self "put - something is wrong";
+            throw $self "- {{INTERNAL:Unknown field type '$type'}}";
         }
     }
 
-    ##
     # Storing all values at once
     #
     if($detached) {
-        @{$$self->{data}}{@data_keys}=values %$data;
+        @{$$self->{'data'}}{@data_keys}=values %$data;
     }
     else {
         $self->_store_data_fields($data);
     }
 
-    ##
     # For multi-value put with hash reference it will return undef,
     # but that's OK -- there is no good answer to what "the value" is
     # anyway.
